#include <security/pam_appl.h>
#include <security/pam_modules.h>
#include <security/pam_misc.h>

#include "authenticator.h"

struct _SessionAuthenticator
{
    BaseObject parent_object;
    gchar* name;
    pam_handle_t *pam;
};

G_DEFINE_FINAL_TYPE(SessionAuthenticator, session_authenticator, BASE_TYPE_OBJECT);

static void session_authenticator_finalize(GObject *object)
{
    g_debug("session_authenticator_finalize: %s", base_object_to_string(object));
    G_OBJECT_CLASS(session_authenticator_parent_class)->finalize(object);
    SessionAuthenticator *self = SESSION_AUTHENTICATOR(object);
    if(self->name)
    {
        g_free(self->name);
    }
    self->name = NULL;
    if(self->pam)
    {
        session_authenticator_close_session(self);
    }
}

static void session_authenticator_dispose(GObject *object)
{
    g_debug("session_authenticator_dispose: %s", base_object_to_string(object));
    G_OBJECT_CLASS(session_authenticator_parent_class)->dispose(object);
}

static const gchar* session_authenticator_to_string(BaseObject *object)
{
    (void)object;
    return "SessionAuthenticator";
}

static void session_authenticator_class_init(SessionAuthenticatorClass *class)
{
    GObjectClass *gobject_class = G_OBJECT_CLASS(class);
    gobject_class->finalize = session_authenticator_finalize;
    gobject_class->dispose = session_authenticator_dispose;

    BaseObjectClass *base_class = BASE_OBJECT_CLASS(class);
    base_class->to_string = session_authenticator_to_string;
}

static void session_authenticator_init(SessionAuthenticator *self)
{
    self->name = NULL;
    self->pam = NULL;
}

static int pam_conversation(int num, const struct pam_message **messages, struct pam_response **responses, void *data)
{
    g_debug("Initiating conversation between PAM and application");

    /* Check number of responses */
    if ((num <= 0) || (num > PAM_MAX_NUM_MSG))
    {
        g_warning("Invalid number of responses '%d'.", num);
        return PAM_CONV_ERR;
    }

    /* Allocate memory for response */
    struct pam_response *resp;

    if (!(resp = calloc(num, sizeof(*resp))))
    {
        return PAM_BUF_ERR;
    }

    /* Iterate over PAM messages */
    char **credentials = (char **)data;
    char *username;
    char *password;
    int i;

    for (i = 0; i < num; i++)
    {
        switch (messages[i]->msg_style)
        {
        case PAM_PROMPT_ECHO_ON:
            username = credentials[0];
            resp[i].resp = strdup(username);

            if (!(resp[i].resp))
            {
                goto fail;
            }

            break;

        case PAM_PROMPT_ECHO_OFF:
            password = credentials[1];
            resp[i].resp = strdup(password);

            if (!(resp[i].resp))
            {
                goto fail;
            }

            break;

        case PAM_TEXT_INFO:
            printf("PAM_TEXT_INFO: %s", messages[i]->msg);
            break;

        case PAM_ERROR_MSG:
            printf("PAM_ERROR_MSG: %s", messages[i]->msg);
            break;

        default:
            goto fail;
        }
    }

    *responses = resp;

    g_debug("PAM conversation completed.");

    return PAM_SUCCESS;

/* Goto failure */
fail:
    g_debug("PAM conversation failed.");

    for (i = 0; i < num; i++)
    {
        if (resp[i].resp)
        {
            memset(resp[i].resp, 0, strlen(resp[i].resp));
            free(resp[i].resp);
        }
    }

    memset(resp, 0, num * sizeof(*resp));
    *responses = 0;

    return PAM_CONV_ERR;
}

SessionAuthenticator* session_authenticator_new(const char* name)
{
    SessionAuthenticator *self = g_object_new(SESSION_TYPE_AUTHENTICATOR, NULL);
    self->name = g_strdup(name);
    return self;
}

gboolean session_authenticator_authenticate(SessionAuthenticator* self, const char* username, const char* password)
{
    if(self->pam)
    {
        g_critical("PAM session is already opened.");
        return FALSE;
    }

    const char *data[2] = {username, password};
    struct pam_conv conversation = {pam_conversation, data};

    /* Start pam */
    g_debug("Starting PAM transaction for user %s", username);

    int result = pam_start(self->name, username, &conversation, &self->pam);

    if (result != PAM_SUCCESS)
    {
        g_critical("Failed to start PAM transaction: %s", pam_strerror(self->pam, result));
        session_authenticator_close_session(self);
        return FALSE;
    }

    /* Set DISPLAY */
    result = pam_set_item(self->pam, PAM_XDISPLAY, getenv("DISPLAY"));

    if (result != PAM_SUCCESS)
    {
        g_warning("Failed to set PAM_XDISPLAY item: %s", pam_strerror(self->pam, result));
    }

    /* Set TTY */
    result = pam_set_item(self->pam, PAM_TTY, getenv("TTY"));

    if (result != PAM_SUCCESS)
    {
        g_warning("Failed to set PAM_TTY item: %s", pam_strerror(self->pam, result));
    }

    /* Set USER */
    result = pam_set_item(self->pam, PAM_USER, username);

    if (result != PAM_SUCCESS)
    {
        g_warning("Failed to set PAM_USER item: %s", pam_strerror(self->pam, result));
    }

    /* Authenticate pam user */
    g_debug("Authenticating username %s", username);

    result = pam_authenticate(self->pam, 0);

    if (result != PAM_SUCCESS)
    {
        g_critical("Failed to authenticate user: %s", pam_strerror(self->pam, result));
        session_authenticator_close_session(self);
        return FALSE;
    }

    /* Check if user account is valid */
    result = pam_acct_mgmt(self->pam, 0);

    if (result != PAM_SUCCESS)
    {
        g_critical("Failed to manage user account: %s", pam_strerror(self->pam, result));
        session_authenticator_close_session(self);
        return FALSE;
    }

    /* Establish credentials */
    result = pam_setcred(self->pam, PAM_ESTABLISH_CRED);

    if (result != PAM_SUCCESS)
    {
        g_critical("Failed to establish PAM credentials: %s", pam_strerror(self->pam, result));
        session_authenticator_close_session(self);
        return FALSE;
    }
    return TRUE;
}

gboolean session_authenticator_is_authenticated(SessionAuthenticator* self)
{
    return self->pam != NULL;
}

gboolean session_authenticator_open_session(SessionAuthenticator* self)
{
    g_debug("Preparing to open PAM session.");
    if(!self->pam)
    {
        g_critical("User unauthorized.");
        return FALSE;
    }

    int result = pam_open_session(self->pam, 0);

    if (result != PAM_SUCCESS) {
        g_critical("Failed to open PAM session: %s", pam_strerror(self->pam, result));
        session_authenticator_close_session(self);
        return FALSE;
    }
    return TRUE;
}

void session_authenticator_close_session(SessionAuthenticator* self)
{
    if(!self->pam)
    {
        g_debug("PAM session is not opened.");
        return;
    }
    g_debug("Ending PAM login session");

    /* Close pam session */
    int result = pam_close_session(self->pam, 0);
    if (result != PAM_SUCCESS)
    {
        g_debug("Failed to close PAM session: %s", pam_strerror(self->pam, result));
    }

    /* Remove credentials */
    result = pam_setcred(self->pam, PAM_DELETE_CRED);

    if (result != PAM_SUCCESS)
    {
        g_debug("Failed to delete PAM credentials: %s", pam_strerror(self->pam, result));
    }

    /* End pam session */
    result = pam_end(self->pam, result);

    if (result != PAM_SUCCESS)
    {
        g_debug("Failed to end PAM session: %s", pam_strerror(self->pam, result));
    }
    self->pam = NULL;
}
